"use strict";

function _typeof(obj) { "@babel/helpers - typeof"; if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") { _typeof = function _typeof(obj) { return typeof obj; }; } else { _typeof = function _typeof(obj) { return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj; }; } return _typeof(obj); }

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.createTransaction = createTransaction;

var _utxoLib = require("@trezor/utxo-lib");

var _bignumber = _interopRequireDefault(require("bignumber.js"));

var _permutation = require("./permutation");

var coinselect = _interopRequireWildcard(require("./coinselect"));

var request = _interopRequireWildcard(require("./request"));

var _bchaddr = require("../utils/bchaddr");

function _getRequireWildcardCache(nodeInterop) { if (typeof WeakMap !== "function") return null; var cacheBabelInterop = new WeakMap(); var cacheNodeInterop = new WeakMap(); return (_getRequireWildcardCache = function _getRequireWildcardCache(nodeInterop) { return nodeInterop ? cacheNodeInterop : cacheBabelInterop; })(nodeInterop); }

function _interopRequireWildcard(obj, nodeInterop) { if (!nodeInterop && obj && obj.__esModule) { return obj; } if (obj === null || _typeof(obj) !== "object" && typeof obj !== "function") { return { "default": obj }; } var cache = _getRequireWildcardCache(nodeInterop); if (cache && cache.has(obj)) { return cache.get(obj); } var newObj = {}; var hasPropertyDescriptor = Object.defineProperty && Object.getOwnPropertyDescriptor; for (var key in obj) { if (key !== "default" && Object.prototype.hasOwnProperty.call(obj, key)) { var desc = hasPropertyDescriptor ? Object.getOwnPropertyDescriptor(obj, key) : null; if (desc && (desc.get || desc.set)) { Object.defineProperty(newObj, key, desc); } else { newObj[key] = obj[key]; } } } newObj["default"] = obj; if (cache) { cache.set(obj, newObj); } return newObj; }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { "default": obj }; }

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) { symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); } keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function _toConsumableArray(arr) { return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _unsupportedIterableToArray(arr) || _nonIterableSpread(); }

function _nonIterableSpread() { throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method."); }

function _unsupportedIterableToArray(o, minLen) { if (!o) return; if (typeof o === "string") return _arrayLikeToArray(o, minLen); var n = Object.prototype.toString.call(o).slice(8, -1); if (n === "Object" && o.constructor) n = o.constructor.name; if (n === "Map" || n === "Set") return Array.from(o); if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen); }

function _iterableToArray(iter) { if (typeof Symbol !== "undefined" && iter[Symbol.iterator] != null || iter["@@iterator"] != null) return Array.from(iter); }

function _arrayWithoutHoles(arr) { if (Array.isArray(arr)) return _arrayLikeToArray(arr); }

function _arrayLikeToArray(arr, len) { if (len == null || len > arr.length) len = arr.length; for (var i = 0, arr2 = new Array(len); i < len; i++) { arr2[i] = arr[i]; } return arr2; }

function inputComparator(aHash, aVout, bHash, bVout) {
  return reverseBuffer(aHash).compare(reverseBuffer(bHash)) || aVout - bVout;
}

function outputComparator(aScript, aValue, bScript, bValue) {
  return new _bignumber["default"](aValue).comparedTo(new _bignumber["default"](bValue)) || aScript.compare(bScript);
} // types for building the transaction in trezor.js


function createTransaction(allInputs, selectedInputs, allOutputs, selectedOutputs, segwit, inputAmounts, basePath, changeId, changeAddress, network, skipPermutation) {
  var convertedInputs = selectedInputs.map(function (input) {
    var id = input.id;
    var richInput = allInputs[id];
    return convertInput(richInput, segwit, inputAmounts, basePath);
  });
  var convertedOutputs = selectedOutputs.map(function (output, i) {
    // change is always last
    var isChange = i === allOutputs.length;
    var original = allOutputs[i]; // null if change

    if (!isChange && original.type === 'opreturn') {
      var opReturnData = original.dataHex;
      return convertOpReturnOutput(opReturnData);
    } // TODO refactor and get rid of FlowIssues everywhere
    // $FlowIssue


    var address = isChange ? changeAddress : original.address;
    var amount = output.value;
    return convertOutput(address, amount, network, basePath, changeId, isChange, segwit);
  }); // this syntax is forced by flow: sketchy-null-bool
  // i don't like it but needs to be refactored anyway...

  if (skipPermutation === true) {
    return {
      inputs: convertedInputs,
      outputs: {
        sorted: convertedOutputs.map(function (o) {
          return o.output;
        })
      }
    };
  }

  convertedInputs.sort(function (a, b) {
    return inputComparator(a.hash, a.index, b.hash, b.index);
  });

  var permutedOutputs = _permutation.Permutation.fromFunction(convertedOutputs, function (a, b) {
    var aValue = typeof a.output.value === 'string' ? a.output.value : '0';
    var bValue = typeof b.output.value === 'string' ? b.output.value : '0';
    return outputComparator(a.script, aValue, b.script, bValue);
  }).map(function (o) {
    return o.output;
  });

  return {
    inputs: convertedInputs,
    outputs: permutedOutputs
  };
}

function convertInput(utxo, segwit, inputAmounts, basePath) {
  var res = {
    hash: reverseBuffer(Buffer.from(utxo.transactionHash, 'hex')),
    index: utxo.index,
    path: basePath.concat(_toConsumableArray(utxo.addressPath)),
    segwit: segwit
  };

  if (inputAmounts) {
    return _objectSpread(_objectSpread({}, res), {}, {
      amount: utxo.value
    });
  }

  return res;
}

function convertOpReturnOutput(opReturnData) {
  var opReturnDataBuffer = Buffer.from(opReturnData, 'hex');
  var output = {
    opReturnData: opReturnDataBuffer
  };

  var script = _utxoLib.script.nullData.output.encode(opReturnDataBuffer);

  return {
    output: output,
    script: script
  };
}

function convertOutput(address, value, network, basePath, changeId, isChange, segwit) {
  var output = isChange ? {
    path: [].concat(_toConsumableArray(basePath), [1, changeId]),
    segwit: segwit,
    value: value
  } : {
    address: address,
    value: value
  };
  return {
    output: output,
    script: _utxoLib.address.toOutputScript((0, _bchaddr.convertCashAddress)(address), network)
  };
}

function reverseBuffer(src) {
  var buffer = Buffer.alloc(src.length);

  for (var i = 0, j = src.length - 1; i <= j; ++i, --j) {
    buffer[i] = src[j];
    buffer[j] = src[i];
  }

  return buffer;
}